/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | www.openfoam.com
     \\/     M anipulation  |
-------------------------------------------------------------------------------
    Copyright (C) 2011-2017 OpenFOAM Foundation
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "headLossPressureFvPatchScalarField.H"
#include "addToRunTimeSelectionTable.H"
#include "fvPatchFieldMapper.H"
#include "volFields.H"
#include "surfaceFields.H"


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::headLossPressureFvPatchScalarField::headLossPressureFvPatchScalarField
(
    const fvPatch& p,
    const DimensionedField<scalar, volMesh>& iF
)
:
    fixedValueFvPatchScalarField(p, iF),
    UName_("U"),
    phiName_("phi"),
    patm_(p.size(), Zero),
    d_(Zero),
    g_(9.81),
    H_(Zero),
    frictionLossFactor_(),
    minorLossFactor_()
{}

Foam::headLossPressureFvPatchScalarField::headLossPressureFvPatchScalarField
(
    const fvPatch& p,
    const DimensionedField<scalar, volMesh>& iF,
    const dictionary& dict
)
:
    fixedValueFvPatchScalarField(p, iF, dict, false),
    UName_(dict.lookupOrDefault<word>("U", "U")),
    phiName_(dict.lookupOrDefault<word>("phi", "phi")),
    patm_("patm", dict, p.size()),
    d_(dict.get<scalar>("d")),
    g_(dict.getOrDefault<scalar>("g",9.81)),
    H_(dict.get<scalar>("H")),
    frictionLossFactor_(dict.lookup("frictionLossFactor")),
    minorLossFactor_(dict.lookup("minorLossFactor"))
{
    if (dict.found("value"))
    {
        fvPatchField<scalar>::operator=
        (
            scalarField("value", dict, p.size())
        );
    }
    else
    {
        fvPatchField<scalar>::operator=(patm_);
    }
}

Foam::headLossPressureFvPatchScalarField::headLossPressureFvPatchScalarField
(
    const headLossPressureFvPatchScalarField& ptf,
    const fvPatch& p,
    const DimensionedField<scalar, volMesh>& iF,
    const fvPatchFieldMapper& mapper
)
:
    fixedValueFvPatchScalarField(ptf, p, iF, mapper),
    UName_(ptf.UName_),
    phiName_(ptf.phiName_),
    patm_(ptf.patm_, mapper),
    d_(ptf.d_),
    g_(ptf.g_),
    H_(ptf.H_),
    frictionLossFactor_(ptf.frictionLossFactor_),
    minorLossFactor_(ptf.minorLossFactor_)
{}


Foam::headLossPressureFvPatchScalarField::headLossPressureFvPatchScalarField
(
    const headLossPressureFvPatchScalarField& tppsf
)
:
    fixedValueFvPatchScalarField(tppsf),
    UName_(tppsf.UName_),
    phiName_(tppsf.phiName_),
    patm_(tppsf.patm_),
    d_(tppsf.d_),
    g_(tppsf.g_),
    H_(tppsf.H_),
    frictionLossFactor_(tppsf.frictionLossFactor_),
    minorLossFactor_(tppsf.minorLossFactor_)
{}


Foam::headLossPressureFvPatchScalarField::headLossPressureFvPatchScalarField
(
    const headLossPressureFvPatchScalarField& tppsf,
    const DimensionedField<scalar, volMesh>& iF
)
:
    fixedValueFvPatchScalarField(tppsf, iF),
    UName_(tppsf.UName_),
    phiName_(tppsf.phiName_),
    patm_(tppsf.patm_),
    d_(tppsf.d_),
    g_(tppsf.g_),
    H_(tppsf.H_),
    frictionLossFactor_(tppsf.frictionLossFactor_),
    minorLossFactor_(tppsf.minorLossFactor_)
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void Foam::headLossPressureFvPatchScalarField::autoMap
(
    const fvPatchFieldMapper& m
)
{
    fixedValueFvPatchScalarField::autoMap(m);
    patm_.autoMap(m);
}


void Foam::headLossPressureFvPatchScalarField::rmap
(
    const fvPatchScalarField& ptf,
    const labelList& addr
)
{
    fixedValueFvPatchScalarField::rmap(ptf, addr);

    const headLossPressureFvPatchScalarField& tiptf =
        refCast<const headLossPressureFvPatchScalarField>(ptf);

    patm_.rmap(tiptf.patm_, addr);
}

double Foam::headLossPressureFvPatchScalarField::fFunc
(
    const scalar& dia,
    const scalar& eps,
    const scalar& Usim,
    const scalar& nu
)
{
    const scalar& U = Usim*sqr(d_/dia);
    const scalar& Re = dia*U/nu;
    
    scalar f(0);

    if ( Re == 0 )
    {
        f = 0;
    }
    else if ( Re < 2300 )
    {
        f = 64/Re;
        Info << "Laminar Flow" << endl;    
        Info << "Re = " << Re << endl;    
    }
    else
    {
        const scalar tol = 1e-5;
        const scalar fInitaL = sqr(1/(-1.8*log10(6.9/Re + pow((eps/dia)/3.7 ,1.11))));
        scalar fOld = fInitaL;
        
        for(int i = 0; i < 20; i++)
        {
            f = sqr(1/(-2*log10((eps/dia)/3.7 + 2.51/(Re*sqrt(fOld)))));
            
            if (mag(f - fOld)/fOld <= tol )
            {
                break;
            }
            fOld = f;
        }
        
        if ( mag(f - fInitaL)/fInitaL >= 0.1 )
        {
            Info << "Using f Inital" << endl;
            f = fInitaL;
        }
    }
    
    return f;
}

void Foam::headLossPressureFvPatchScalarField::updateCoeffs
(
    const scalarField& patmp,
    const vectorField& Up
)
{
    if (updated())
    {
        return;
    }

    const fvsPatchField<scalar>& phip =
        patch().lookupPatchField<surfaceScalarField, scalar>(phiName_);
             
      
    // Calculating the Average velocity on the patch, Uavg = phi/A
    scalar sumPhi = gSum(patch().lookupPatchField<surfaceScalarField, scalar>(phiName_));
    const scalar& totalArea = gSum(patch().magSf());
    scalar Uavg;
    if (sumPhi != 0)
    {
        Uavg = mag(sumPhi)/totalArea;
    }
    else
    {
        Info << "Zero flux on patch" << endl;
        sumPhi = 1e-20;
        Uavg = mag(sumPhi)/totalArea;
    }
 
    // Calculating the Minor losses
    scalar dpMinor(0);
    const scalar * dMinor;
    const scalar * k;
    for(int i = 0; i < minorLossFactor_.size(); i++)
    {
        dMinor = &minorLossFactor_[i].first().component(0);
        k = &minorLossFactor_[i].first().component(1);
        dpMinor = *k * sqr(Uavg * sqr(d_ / *dMinor))/2 + dpMinor;
    }
    
    // Reading Kinematic viscosity from transportProperties
    const dictionary& transportProperties=
        db().lookupObject<dictionary>("transportProperties");
    const dimensionedScalar nuDimScal("nu", dimViscosity, transportProperties);
    const scalar& nu = nuDimScal.value();

    // Calculating the Friction Losses
    scalar f(0);
    scalar dpFriction(0);
    const scalar * dFriction;
    const scalar * epsilon;
    const scalar * L;
    for(int i = 0; i < frictionLossFactor_.size(); i++)
    {    
        dFriction = &frictionLossFactor_[i].first().component(0);
        epsilon = &frictionLossFactor_[i].first().component(1);
        L = &frictionLossFactor_[i].first().component(2);
        
        f = fFunc(*dFriction,*epsilon,Uavg,nu);
        dpFriction = f * *L / *dFriction * sqr(Uavg * sqr(d_  / *dFriction))/2 + dpFriction;
    }
        
    if (internalField().dimensions() == dimPressure/dimDensity)
    {
        // Incompressible flow
        operator==
        (
            patmp - 0.5*(1.0 - pos0(phip))*magSqr(Up)
             + sumPhi/mag(sumPhi)*(dpFriction + dpMinor) + g_*H_
        );
    }
    else
    {
        FatalErrorInFunction
            << " Incorrect pressure dimensions " << internalField().dimensions()
            << nl
            << "    Should be " << dimPressure
            << " for compressible/variable density flow" << nl
            << "    or " << dimPressure/dimDensity
            << " for incompressible flow," << nl
            << "    on patch " << this->patch().name()
            << " of field " << this->internalField().name()
            << " in file " << this->internalField().objectPath()
            << exit(FatalError);
    }

    fixedValueFvPatchScalarField::updateCoeffs();
}


void Foam::headLossPressureFvPatchScalarField::updateCoeffs()
{
    updateCoeffs
    (
        patm(),
        patch().lookupPatchField<volVectorField, vector>(UName())
    );
}


void Foam::headLossPressureFvPatchScalarField::write(Ostream& os) const
{
    fvPatchScalarField::write(os);
    os.writeEntryIfDifferent<word>("U", "U", UName_);
    os.writeEntryIfDifferent<word>("phi", "phi", phiName_);
    os.writeEntry<scalar>("d", d_);
    os.writeEntryIfDifferent<scalar>("g", 9.81,g_);
    os.writeEntry<scalar>("H", H_);
    os.writeEntry("frictionLossFactor",frictionLossFactor_);
    os.writeEntry("minorLossFactor",minorLossFactor_);
    patm_.writeEntry("patm", os);
    writeEntry("value", os);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
    makePatchTypeField
    (
        fvPatchScalarField,
        headLossPressureFvPatchScalarField
    );
}

// ************************************************************************* //
