/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | www.openfoam.com
     \\/     M anipulation  |
-------------------------------------------------------------------------------
    Copyright (C) 2011-2016 OpenFOAM Foundation
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::headLossPressureFvPatchScalarField

Group
    grpInletBoundaryConditions grpOutletBoundaryConditions

Description
    This boundary condition provides an incompressible total pressure condition for an inflow patch 
    and a static pressure condition for outlow patch with head losses and surface elevation included.
    
    1. incompressible total pressure for inflow patch:
        \f[
            p_p = p_atm - 0.5 |U|^2 - dp_loss + g*H
        \f]
    2. incompressible static pressure for outflow patch:
        \f[
            p_p = p_atm + dp_loss + g*H
        \f]
        where
        \vartable
            p_p     | incompressible pressure at patch [m2/s2]
            p_atm   | incompressible total pressure in the reservior [m2/s2]
            U       | velocity [m/s]
            dp_loss | pressure losses due to minor and friction factors [m2/s2]
            H       | Elevation of the reservior [m]
            g       | Gravity [kg m/s2]
        \endvartable

	The loss factor dp_loss is based on steady state loss factors and calculated as:
	    \f[
	    	dp_loss = sum(f_i*L_i/d_i * 0.5 |U_i|^2) + sum(k_j *0.5 |U_j|^2)
	    \f]
	Where f is the friction factor, L is pipe length, d is diameter of the pipe,
	k is the loss coefficient. f is directly solved in a laminar case as:
	    \f[
	    	f = 64/Re
	    \f]
	and iteratively via the Colebrook equation in a turbulent case as:
	    \f[
	    	1/sqrt(f) = -2.0 log10( (epsilon/d)/3.7 + 2.51/(Re*sqrt(f)))
	    \f[
	Where epsilon is the surface roughness and Re the Reynolds number. The swich 
	between turbulent and laminar is assumed to be at Re = 2300, below laminar
	and above turbulent.


Usage
    \table
        Property        | Description                 | Required | Default value
        U               | Velocity field name         | no       | U
        phi             | Flux field name             | no       | phi
        H               | Level of reservior surface  | no       | 
        g               | Gravity                     | no       | 9.81 
        patm            | Atmospheric pressure        | yes      | 
        d               | Hydralic Diamter of patch   | yes      | 
        frictionLossFactor | Fricition loss factors   | yes      | 
        minorLossFactor | Minor loss factors          | yes      |  
        
    \endtable

    Example of the boundary condition specification:
    \verbatim
    <patchName>
    {
        type            headLossPressure;
        patm            uniform 0;  // m2/s2
        H               10;         // m
        d               1;          // m
        frictionLossFactor
        (
            (($d 2e-05 8) pipe1)
            (($d 2e-05 5) pipe2)
        );
        minorLossFactor
        (
            (($d 0.05) bend)
            (($d 0.01) valve)
        );
    }
    
    The input of frictionLossFactor is based on a Tuple2<vector, word> and the components
    correspond to the pipe hydralic diameter [m], surface rougness (epsilon) [m], pipe length [m],
    and name of the pipe e.g:
    frictionLossFactor = 
        (( diameter_1, epsilon_1, L_1 ) pipe1 )
        (( diameter_2, epsilon_2, L_2 ) pipd2 )
    				  
    
    The input of minorLossFactor is based on a Tuple2<vector2D, word> and the components
    correspond to the pipe diameter [m], loss factor (k), and name of the loss e.g:
    minorLossFactor = 
        (( diameter_1, k_1 ) valve)
        (( diameter_2, k_2 ) bend)
    				  
    All parameters must be supplied in SI units, i.e. [m].  				  
    				  
    \endverbatim    

See also
    Foam::fixedValueFvPatchField

SourceFiles
    headLossPressureFvPatchScalarField.C

\*---------------------------------------------------------------------------*/

#ifndef headLossPressureFvPatchScalarField_H
#define headLossPressureFvPatchScalarField_H

#include "fixedValueFvPatchFields.H"
#include "vector2DField.H"
#include "Tuple2.H"



// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
              Class headLossPressureFvPatchScalarField Declaration
\*---------------------------------------------------------------------------*/

class headLossPressureFvPatchScalarField
:
    public fixedValueFvPatchScalarField
{
    // Private data

        //- Name of the velocity field
        word UName_;

        //- Name of the flux transporting the field
        word phiName_;

        //- Atmospheric pressure
        scalarField patm_;
            
        //- Hydraulic diameter of the pipe
        const scalar d_;

        //- Gravity
        const scalar g_;
        
        //- Height
        const scalar H_;        
                      
        //- Tuple2 list of the friction loss factors ((d, epsilon, L) name)
        typedef Tuple2<vector, word> indexedVector;      
        const List<indexedVector> frictionLossFactor_;
        
        //- Tuple2 list of the minor loss factors ((d, k) name)
        typedef Tuple2<vector2D, word> indexedVector2D;      
        const List<indexedVector2D> minorLossFactor_;

	
public:

    //- Runtime type information
    TypeName("headLossPressure");


    // Constructors

        //- Construct from patch and internal field
        headLossPressureFvPatchScalarField
        (
            const fvPatch&,
            const DimensionedField<scalar, volMesh>&
        );

        //- Construct from patch, internal field and dictionary
        headLossPressureFvPatchScalarField
        (
            const fvPatch&,
            const DimensionedField<scalar, volMesh>&,
            const dictionary&
        );

        //- Construct by mapping given headLossPressureFvPatchScalarField
        //  onto a new patch
        headLossPressureFvPatchScalarField
        (
            const headLossPressureFvPatchScalarField&,
            const fvPatch&,
            const DimensionedField<scalar, volMesh>&,
            const fvPatchFieldMapper&
        );

        //- Construct as copy
        headLossPressureFvPatchScalarField
        (
            const headLossPressureFvPatchScalarField&
        );

        //- Construct and return a clone
        virtual tmp<fvPatchScalarField> clone() const
        {
            return tmp<fvPatchScalarField>
            (
                new headLossPressureFvPatchScalarField(*this)
            );
        }

        //- Construct as copy setting internal field reference
        headLossPressureFvPatchScalarField
        (
            const headLossPressureFvPatchScalarField&,
            const DimensionedField<scalar, volMesh>&
        );

        //- Construct and return a clone setting internal field reference
        virtual tmp<fvPatchScalarField> clone
        (
            const DimensionedField<scalar, volMesh>& iF
        ) const
        {
            return tmp<fvPatchScalarField>
            (
                new headLossPressureFvPatchScalarField(*this, iF)
            );
        }


    // Member functions

        // Access

            //- Return the name of the velocity field
            const word& UName() const
            {
                return UName_;
            }

            //- Return reference to the name of the velocity field
            //  to allow adjustment
            word& UName()
            {
                return UName_;
            }

            //- Return the name of the flux field
            const word& phiName() const
            {
                return phiName_;
            }

            //- Return reference to the name of the flux field
            //  to allow adjustment
            word& phiName()
            {
                return phiName_;
            }

            // Return the friction loss coefficient f
            double fFunc
            (
                const scalar&,
                const scalar&,
                const scalar&,
                const scalar&
            );
			            
            //- Return the total pressure
            const scalarField& patm() const
            {
                return patm_;
            }

            //- Return reference to the total pressure to allow adjustment
            scalarField& patm()
            {
                return patm_;
            }


        // Mapping functions

            //- Map (and resize as needed) from self given a mapping object
            virtual void autoMap
            (
                const fvPatchFieldMapper&
            );

            //- Reverse map the given fvPatchField onto this fvPatchField
            virtual void rmap
            (
                const fvPatchScalarField&,
                const labelList&
            );


        // Evaluation functions

            //- Inherit updateCoeffs from fixedValueFvPatchScalarField
            using fixedValueFvPatchScalarField::updateCoeffs;

            //- Update the coefficients associated with the patch field
            //  using the given patch total pressure and velocity fields
            virtual void updateCoeffs
            (
                const scalarField& patmp,
                const vectorField& Up
            );
          
            //- Update the coefficients associated with the patch field
            virtual void updateCoeffs();


        //- Write
        virtual void write(Ostream&) const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
